/*
  mp3tool
  
  Command line tool for moving a MP3 file according to MP3
  information.
  
  Released under MIT license terms, please read LICENSE file.
  
  Bugs or comments can be sent to wpitchoune@gmail.com
 */

/* to use the GNU version of basename */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

/* Directory separator is \ when cross-compiling for MS Windows
   systems */
#if defined(__MINGW32__)
#define DIRSEP '\\'
#else
#define DIRSEP '/'
#endif

#ifndef MP3TOOL_VERSION
#define MP3TOOL_VERSION "0.0.0"
#endif

#define CMD_NAME "mp3tool"

#define CMD_USAGE "mp3tool - v%s by wpitchoune@gmail.com\
\nReleased under MIT license terms, read LICENSE file\
\nhttp://wpitchoune.net/mp3tool\
\n\
\nUsage: mp3tool [OPTION] [mp3_filepath] [pattern]\
\n\
\nOptions: \
\n\t -i\t\t Displays information of the MP3 file according to the 'pattern'\
\n\t -m\t\t Move MP3 file to 'pattern'\
\n\t -h\t\t Displays help\
\n\t -v\t\t Displays version information\
\n\
\nPattern variables:\
\n%%f\tOriginal filename\
\n%%c\tMP3 comment\
\n%%a\tMP3 artist\
\n%%l\tMP3 album\
\n%%t\tMP3 title\
\n%%%%\tPercent\
\n%%n\tNewline\n"

struct mp3_info {
	char title[31];
	char artist[31];
	char album[31];
	char year[5];
	char comment[31];
	char genre;
	char *filename;
};

/*
  Normalize an information field.
 
  If the string only contains whitespaces the string is replaced by
  'Unknown' string.
 */
void normalize(char *str)
{
	int unknown = 1;

	char *c = str;
	while (*c) {
		if (*c != ' ') {
			unknown = 0;
			break;
		}
		c++;
	}

	if (unknown)
		strcpy(str, "Unknown");
}

int read_mp3_info(const char *fpath, struct mp3_info *info)
{
	FILE *file;
	char tag[3];

	info->filename = strdup(basename(fpath));

	file = fopen(fpath, "r");

	if (!file)
		return 0;

	if (fseek(file, -128, SEEK_END) != 0)
		return 0;

	if (fread(&tag, 1, 3, file) != 3
	    || tag[0] != 'T' || tag[1] != 'A' || tag[2] != 'G')
		return 0;

	if (fread(&info->title, 1, 30, file) != 30)
		return 0;
	info->title[30] = '\0';
	normalize(info->title);

	if (fread(&info->artist, 1, 30, file) != 30)
		return 0;
	info->artist[30] = '\0';
	normalize(info->artist);

	if (fread(&info->album, 1, 30, file) != 30)
		return 0;
	info->album[30] = '\0';
	normalize(info->album);

	if (fread(&info->year, 1, 4, file) != 4)
		return 0;
	info->year[4] = '\0';

	if (fread(&info->comment, 1, 30, file) != 30)
		return 0;
	info->comment[30] = '\0';

	if (fread(&info->genre, 1, 1, file) != 1)
		return 0;

	fclose(file);

	return 1;
}

char *apply_pattern(const char *pattern, struct mp3_info *info)
{
	char *result = malloc(1);
	char *p;

	*result = '\0';
	p = (char *)pattern;

	while (*p) {
		char *old = result;
		int n;

		if (*p == '%' && *(p + 1)) {
			char *str = NULL;
			switch (*(p + 1)) {
			case 'a':
				str = info->artist;
				break;
			case 'l':
				str = info->album;
				break;
			case 'f':
				str = info->filename;
				break;
			case 't':
				str = info->title;
				break;
			case 'c':
				str = info->comment;
				break;
			case 'y':
				str = info->year;
				break;
			case 'n':
				str = "\n";
				break;
			case '%':
				str = "%";
				break;
			}

			if (str) {
				n = strlen(old) + strlen(str) + 1;
				result = malloc(n);
				strcpy(result, old);
				strcat(result, str);
				p++;
			} else {
				old = NULL;
			}
		} else {
			n = strlen(old) + 1 + 1;

			result = malloc(n);

			strcpy(result, old);
			result[n - 2] = *p;
			result[n - 1] = '\0';
		}

		free(old);

		p++;
	}

	return result;
}

void mkdirs(const char *dirs)
{
	char *c = (char *)dirs;
	char *dir = malloc(strlen(dirs) + 1);

	int i = 0;
	while (*c) {
		if ((*c == DIRSEP || *c == '\0') && c != dirs) {
			strncpy(dir, dirs, i);
			dir[i] = '\0';
			mkdir(dir, 0777);
		}

		c++;
		i++;
	}
	free(dir);

	mkdir(dirs, 0777);
}

char *wp_dirname(const char *path)
{
	int n = strlen(path);

	while (n >= 0) {
		if (path[n] == DIRSEP) {
			char *res = strdup(path);
			res[n] = '\0';
			return res;
		}

		n--;
	}

	return NULL;
}

int mp3_move(const char *fpath, const char *newpath_pattern)
{
	struct mp3_info info;
	char *npath, *dir;
	int ret = read_mp3_info(fpath, &info);

	if (!ret)
		return 0;

	npath = apply_pattern(newpath_pattern, &info);

	dir = wp_dirname(npath);

	if (dir)
		mkdirs(dir);

	printf("move: %s to %s\n", fpath, npath);
	ret = rename(fpath, npath);

	free(npath);
	free(dir);
	free(info.filename);

	return ret == 0;
}

int main(int argc, char **argv)
{
	if (argc == 4 && !strcmp(argv[1], "-i")) {
		char *res, *fpath;
		struct mp3_info info;

		fpath = argv[2];

		if (!read_mp3_info(fpath, &info)) {
			fprintf(stderr,
				"ERROR: Failure to read MP3 information %s\n",
				fpath);
			exit(EXIT_FAILURE);
		}
		res = apply_pattern(argv[3], &info);
		printf("%s", res);
		free(res);
		free(info.filename);

	} else if (argc == 4 && !strcmp(argv[1], "-m")) {
		char *fpath = argv[2];

		if (!mp3_move(fpath, argv[3])) {
			fprintf(stderr, "ERROR: Cannot move %s\n", fpath);
			exit(EXIT_FAILURE);
		}

	} else if (argc == 2 && !strcmp(argv[1], "-h")) {
		printf(CMD_USAGE, MP3TOOL_VERSION);

	} else if (argc == 2 && !strcmp(argv[1], "-v")) {
		printf("%s - %s\n", CMD_NAME, MP3TOOL_VERSION);

	} else {
		fprintf(stderr, CMD_USAGE, MP3TOOL_VERSION);
		exit(EXIT_FAILURE);
	}

	exit(EXIT_SUCCESS);
}
